/*====================================+=====================================+
! File sc_install.cpp                 ! Copyright (C) 2016-2019 Remi PASCAL !
+-------------------------------------+-------------------------------------+
! This is free software: you can redistribute it and/or modify it under the !
! terms of the GNU General Public License as published by the Free Software !
! Foundation, either version 3 of the License, or any later version.        !
! This application is distributed in the hope that it will be useful, but   !
! WITHOUT ANY WARRANTY; without even the implied warranty of                !
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General !
! Public License for more details.                                          !
! You should have received a copy of the GNU General Public License along   !
! with this application. If not, see <http://www.gnu.org/licenses/>.        !
+==========================================================================*/



/*-------------------------------------------------------------------------*/
#include <memory>
#include <wx/sstream.h>
#include <wx/txtstrm.h>
#include <wx/wfstream.h>
#include <wx/zipstrm.h>
#include <wx/filename.h>
#include "app_info.h"
#include "CApp.h"
#include "common/sc_install.h"
/*-------------------------------------------------------------------------*/



/*-------------------------------------------------------------------------*/
constexpr const char *sc::CInstallThread::m_st_a_err_msg[ IE_NB ] ;
/*-------------------------------------------------------------------------*/



/*-------------------------------------------------------------------------*/
wxDEFINE_EVENT( sc::EVT_INSTALL, sc::CInstallEvent ) ;

/*-------------------------------------------------------------------------*/
void sc::CInstallThread::send_event()
{
   /*----------------------------------------------------------------------*/
   CInstallEvent event( EVT_INSTALL, wxID_ANY ) ;
   /*----------------------------------------------------------------------*/
   event.set_err_code( m_err_code ) ;
   event.set_num_file( m_i_num_file ) ;
   event.set_nb_file( m_i_nb_file ) ;
   /*----------------------------------------------------------------------*/
   if( wxThread::IsMain() )
   {  m_p_inst->ProcessEvent( event ) ; }
   else
   {  m_p_inst->QueueEvent( new CInstallEvent( event ) ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
sc::CInstallThread::CInstallThread( CDlgInstall *p_inst )
                  : wxThread( wxTHREAD_DETACHED ),
                    m_p_inst( p_inst ), m_fn_source( p_inst->m_fn_source )
{  ; }

/*-------------------------------------------------------------------------*/
sc::CInstallThread::~CInstallThread()
{
   /*----------------------------------------------------------------------*/
   if( m_p_inst != nullptr )
   {  wxCriticalSectionLocker enter_csec( m_p_inst->m_csec_p_thread ) ;
      m_p_inst->m_p_thread = nullptr ;
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
sc::CInstallThread::e_err_code sc::CInstallThread::install()
{
   /*-----------------------------------------------------------------------+
   ! By default the extraction is done in the executable's directory :      !
   ! the "root" of the application                                          !
   +-----------------------------------------------------------------------*/
   wxString s_path = wxGetApp().get_exe_path() ;
   wxString s_zip  = m_fn_source.GetFullPath() ;

   /*--( This block is here to close the zip and be able to remove it )----*/
   {
      /*--( Try to open the zip file )-------------------------------------*/
      wxFileInputStream fis( s_zip ) ;
      if( !fis.IsOk() ) { return m_err_code = IE_FILE_NOT_FOUND ; }
      /*-------------------------------------------------------------------*/
      wxZipInputStream zis( fis ) ;
      if( !zis.IsOk() ) { return m_err_code = IE_INVALID_ZIP ; }

      /*-------------------------------------------------------------------*/
      m_err_code = IE_NO_ERROR_INFO ;
      m_i_nb_file = zis.GetTotalEntries() ;
      m_i_num_file = 0 ;
      /*-------------------------------------------------------------------*/
      std::unique_ptr< wxZipEntry > up_zentry ;
      wxFileName fno ;

      /*--( The first element is supposed to be the executable directory )-*/
      up_zentry.reset( zis.GetNextEntry() ) ;
      if( up_zentry.get() == nullptr || !up_zentry->IsDir() )
      {  return m_err_code = IE_UNEXPECTED_CONTENT ; }
      /*--( Show this directory as "treated" )-----------------------------*/
      ++m_i_num_file ;
      send_event() ;
      /*--( This directory will be removed from all extracted file names )-*/
      wxString s_main_path = up_zentry->GetName() ;
      wxString s_file_path ;
      /*-------------------------------------------------------------------*/
      const wxString s_exe_name = wxGetApp().get_exe_fullname() ;
      /*-------------------------------------------------------------------*/
      for( up_zentry.reset( zis.GetNextEntry() ) ;
           up_zentry.get() != nullptr ;
           up_zentry.reset( zis.GetNextEntry() )
         )
      {
         /*--( Check if stop requested )-----------------------------------*/
         if( TestDestroy() ) { return m_err_code = IE_STOP_REQUEST ; }

         /*--( All files are supposed to be in this directory )------------*/
         if( !up_zentry->GetName().StartsWith( s_main_path, &s_file_path ) )
         {  return m_err_code = IE_UNEXPECTED_CONTENT ; }
         /*--( The entry name can contain a relative path too )------------*/
         fno.Assign( s_path + wxFileName::GetPathSeparator() + s_file_path );

         /*--( Create directory if it is one )-----------------------------*/
         if( up_zentry->IsDir() )
         {  /*-------------------------------------------------------------*/
            fno.Mkdir( up_zentry->GetMode(), wxPATH_MKDIR_FULL ) ;
            wxDateTime dt_mod = up_zentry->GetDateTime() ;
            fno.SetTimes( nullptr, &dt_mod, nullptr ) ;
            fno.SetPermissions( up_zentry->GetMode() ) ;
            /*-------------------------------------------------------------*/
            ++m_i_num_file ;
            send_event() ;
            /*-------------------------------------------------------------*/
            continue ;
            /*-------------------------------------------------------------*/
         }

         /*-----------------------------------------------------------------+
         ! The executable currently running can't be overwritten            !
         +-----------------------------------------------------------------*/
         if( fno.GetFullName() == s_exe_name )
         {  /*--( So, it is extracted with a different name )--------------*/
            fno.SetName( fno.GetName() + "_new" ) ;
            /*--( And it is executed after exiting this one )--------------*/
            wxGetApp().set_program_to_start_on_exit( fno.GetFullPath() ) ;
            /*-------------------------------------------------------------*/
         }

         /*--( Regular file that is going to be extracted )----------------*/
         zis.OpenEntry( *up_zentry.get() ) ;
         if( !zis.CanRead() ) { return m_err_code = IE_READ_ERROR ; }

         /*--( Create copy )-----------------------------------------------*/
         wxFileOutputStream fos( fno.GetFullPath() ) ;
         if( !fos.IsOk() ) { return m_err_code = IE_WRITE_ERROR ; }
         /*--( Read entry content into the output stream )-----------------*/
         zis.Read( fos ) ;
         /*----------------------------------------------------------------*/
         wxDateTime dt_mod = up_zentry->GetDateTime() ;
         fno.SetTimes( nullptr, &dt_mod, nullptr ) ;
         fno.SetPermissions( up_zentry->GetMode() ) ;
         /*--( Information )-----------------------------------------------*/
         ++m_i_num_file ;
         send_event() ;
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }

   /*--( The installation went well, the zip can be removed )--------------*/
   wxRemove( s_zip ) ;
   /*----------------------------------------------------------------------*/
   return m_err_code = IE_NO_ERROR_DONE ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxThread::ExitCode sc::CInstallThread::Entry()
{
   /*----------------------------------------------------------------------*/
   install() ;
   /*--( Last event )------------------------------------------------------*/
   send_event() ;
   /*----------------------------------------------------------------------*/
   return ( wxThread::ExitCode )0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
sc::CDlgInstall::CDlgInstall( wxWindow *parent, const wxFileName &fn_source )
               : CDlgProgress( parent ),
                 m_fn_source( fn_source )
{
   /*----------------------------------------------------------------------*/
   SetTitle( _( "Installation of" ) + ' ' + fn_source.GetFullName() ) ;
   /*----------------------------------------------------------------------*/
   sc::init_tlw_rect( this, wxGetApp().M_rec_progress ) ;
   Bind( EVT_INSTALL, &CDlgInstall::OnEvent, this ) ;
   Layout() ;
   /*----------------------------------------------------------------------*/
   thread_start() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
sc::CDlgInstall::~CDlgInstall()
{
   /*----------------------------------------------------------------------*/
   sc::save_tlw_rect( this, wxGetApp().M_rec_progress ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void sc::CDlgInstall::OnBtCancel( wxCommandEvent &event )
{
   /*----------------------------------------------------------------------*/
   if( thread_running() )
   {  thread_stop() ; }
   else
   {  /*--------------------------------------------------------------------+
      ! If thread is not running any more this means that the process went  !
      ! well and that the user clicked "ok" to restart                      !
      +--------------------------------------------------------------------*/
      EndModal( CInstallThread::IE_NO_ERROR_DONE ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*--( This function is called from outside of this module )----------------*/
sc::CInstallThread::e_err_code sc::CDlgInstall::non_threaded_exec()
{
   /*----------------------------------------------------------------------*/
   return CInstallThread( this ).install() ;
   /*----------------------------------------------------------------------*/
}

/*--( This function is called from outside of this module )----------------*/
void sc::CDlgInstall::thread_start()
{
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec( m_csec_p_thread ) ;
   /*--( Clean current before new one )------------------------------------*/
   if( m_p_thread != nullptr ) { m_p_thread->reset_p_inst() ; }
   /*----------------------------------------------------------------------*/
   m_p_thread = new CInstallThread( this ) ;
   /*----------------------------------------------------------------------*/
   if( m_p_thread->Create() != wxTHREAD_NO_ERROR )
   {  wxFAIL ; return ; }
   /*----------------------------------------------------------------------*/
   if( m_p_thread->Run() != wxTHREAD_NO_ERROR )
   {  wxFAIL ; return ; }
   /*----------------------------------------------------------------------*/
}

/*--( This function is called from outside of this module )----------------*/
void sc::CDlgInstall::thread_stop()
{
   /*----------------------------------------------------------------------*/
   m_boo_thread_stopping = true ;
   /*----------------------------------------------------------------------*/
   {  wxCriticalSectionLocker enter_csec( m_csec_p_thread ) ;
      /*-------------------------------------------------------------------*/
      if( m_p_thread != nullptr && m_p_thread->Delete() != wxTHREAD_NO_ERROR)
      {  wxFAIL ; return ; }
      /*-------------------------------------------------------------------*/
   }
   /*--( Wait until thread real death ! )----------------------------------*/
   while( true )
   {  /*-------------------------------------------------------------------*/
      {  wxCriticalSectionLocker enter_csec( m_csec_p_thread ) ;
         if( m_p_thread == nullptr ) { break ; }
      }
      /*-------------------------------------------------------------------*/
      wxThread::This()->Sleep( 100 ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   m_boo_thread_stopping = false ;
   /*----------------------------------------------------------------------*/
}

/*--( This function is called from outside of this module )----------------*/
bool sc::CDlgInstall::thread_running()
{
   /*----------------------------------------------------------------------*/
   wxCriticalSectionLocker enter_csec( m_csec_p_thread ) ;
   /*----------------------------------------------------------------------*/
   return m_p_thread != nullptr && !m_boo_thread_stopping ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void sc::CDlgInstall::OnEvent( CInstallEvent &event )
{
   /*----------------------------------------------------------------------*/
   if( event.get_err_code() == CInstallThread::IE_NO_ERROR_INFO )
   {
      /*--( Not too much information ? )-----------------------------------*/
      if( !check_progress_display() ) { return ; }
      /*-------------------------------------------------------------------*/
      wxString msg( wxString::Format( "%d/%d",
                                      event.get_num_file(),
                                      event.get_nb_file()
                                    )
                  ) ;
      /*-------------------------------------------------------------------*/
      msg_info( msg ) ;
      m_ga_progress->SetRange( event.get_nb_file() ) ;
      m_ga_progress->SetValue( event.get_num_file() ) ;
      /*-------------------------------------------------------------------*/
   }
   else
   if( event.get_err_code() == CInstallThread::IE_STOP_REQUEST )
   {  EndModal( event.get_err_code() ) ; }
   else
   if( event.get_err_code() == CInstallThread::IE_NO_ERROR_DONE )
   {
      /*--( On the end the "progress bar" has to be full ... )-------------*/
      m_ga_progress->SetRange( 1 ) ;
      m_ga_progress->SetValue( 1 ) ;
      /*-------------------------------------------------------------------*/
      msg_info( _( "Installation done. Restart needed." ) ) ;
      m_bt_quit->SetLabel( _( "Restart" ) + ' ' + APP_DISPLAY_NAME ) ;
      m_bsz_gen->Layout() ;
      /*-------------------------------------------------------------------*/
   }
   else
   {  msg_error( CInstallThread::st_get_err_msg( event.get_err_code() ) ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void sc::CDlgInstall::msg_info( const wxString &s )
{
   /*----------------------------------------------------------------------*/
   m_stx_msg->SetForegroundColour( *wxBLUE ) ;
   m_stx_msg->SetLabelText( s ) ;
   m_stx_msg->Update() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void sc::CDlgInstall::msg_error( const wxString &s )
{
   /*----------------------------------------------------------------------*/
   m_stx_msg->SetForegroundColour( *wxRED ) ;
   m_stx_msg->SetLabelText( s ) ;
   m_stx_msg->Update() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int sc::CDlgInstall::go()
{
   /*----------------------------------------------------------------------*/
   return ShowModal() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/



/*==========================================================================+
!                      End of file sc_install.cpp                           !
+==========================================================================*/
